#define PROLOG_MODULE "protorch"
#include <SWI-Prolog.h>
#include <SWI-cpp.h>

#include <torch/script.h>

#include <iostream>
#include <iostream>
#include <fstream>
#include <memory>
#include <cassert>
#include <cmath>
#include <cstdlib>
#include <cstdio>

using std::size_t;
using std::pair;
using std::vector;
using std::unordered_map;
using std::string;
using std::make_pair;
using std::endl;
using std::cout;
using std::cerr;

typedef at::Tensor Tensor;
typedef torch::jit::IValue IVal;
typedef std::shared_ptr<torch::jit::script::Module> Model;

string modeldir;

PREDICATE(init, 1)
{
  modeldir = (char *)A1;
  
  cout << "Set modeldir to " << modeldir << endl;
  
  return TRUE;
}

static unordered_map<string,long> load_constant_id_lookup()
{
  std::ifstream infile(modeldir+"/translate_const.txt");
  
  unordered_map<string,long> result;
  
  string sym;
  long id;
  while (infile >> sym >> id)
  {
    result[sym] = id;
  }
  return result;
}

static Tensor embed_constant(const string& name)
{
  static unordered_map<string,long> constant_id_lookup = load_constant_id_lookup();
  static unordered_map<string,Tensor> constant_embeddings;

  auto search_name = constant_embeddings.find(name);
  if (search_name != constant_embeddings.end()) {
    return search_name->second;
  }
  else
  {
    static Model m = torch::jit::load(modeldir+"/s_emb.pt");
    
    auto search_id = constant_id_lookup.find(name);
    if (search_id == constant_id_lookup.end()) {
      cerr << "Couldn't resolve constant index for "<< name << endl;
      exit(1);
    }
    
    static std::vector<IVal> inputs;
    inputs.clear();
    inputs.push_back(torch::tensor((int64_t)search_id->second,at::kLong));
    Tensor result = m->forward(inputs).toTensor()[0];
    constant_embeddings[name] = result;
  
    return result;
  }
}

static Model get_named_model(const string& name)
{
  static unordered_map<string,Model> named_model_cache;
  
  auto search = named_model_cache.find(name);
  if (search != named_model_cache.end()) {
    return search->second;
  } else {
    Model m = torch::jit::load(modeldir+"/str/"+name+".pt");
    named_model_cache[name] = m;
    return m;
  }
}

static Tensor embed_term(PlTerm t)
{
  switch (t.type()) {
    case PL_VARIABLE:
      return embed_constant("X1");
    case PL_ATOM:
    case PL_INTEGER:
    case PL_FLOAT:
      return embed_constant((char *)t);
    case PL_TERM: {
      vector<Tensor> sub_embs; // must be non-static, because of recursion
      int arity = t.arity();
      for (int i = 1; i<=arity; i++) {
        sub_embs.push_back(embed_term(t[i]));
      }
      Model m = get_named_model(t.name());
      static std::vector<IVal> inputs;
      inputs.clear();
      inputs.push_back(torch::cat(sub_embs,0));
      return m->forward(inputs).toTensor();
    }
    default: { // assuming anything else is a list!
      PlTail tail(t);
      PlTerm e;
      vector<Tensor> sub_embs; // must be non-static, because of recursion
      while(tail.next(e)) {
        sub_embs.push_back(embed_term(e));
      }
      static Model m = torch::jit::load(modeldir+"/clausenet.pt");
      static std::vector<IVal> inputs;
      inputs.clear();
      inputs.push_back(torch::cat(sub_embs,0));
      return m->forward(inputs).toTensor();
    }
  }
}

static Tensor embed_path(PlTerm t)
{
  // assuming a list
  PlTail tail(t);
  PlTerm e;
  
  static vector<Tensor> sub_embs; // can be static, path is only toplevel
  sub_embs.clear();
  while(tail.next(e)) {
    sub_embs.push_back(embed_term(e));
  }
  
  static Model m = torch::jit::load(modeldir+"/conjecturenet.pt");
  static std::vector<IVal> inputs;
  inputs.clear();
  inputs.push_back(torch::cat(sub_embs,0));
  return m->forward(inputs).toTensor();
}

PREDICATE(eval_path_term, 4)
{
  static Model m = torch::jit::load(modeldir+"/final.pt");
  
  Tensor t1 = embed_path(A1);
  Tensor t2 = embed_term(A2);
  
  static vector<IVal> inputs;
  inputs.clear();
  inputs.push_back(torch::cat({t1,t2},0));
  Tensor output = m->forward(inputs).toTensor();
  
  auto raw_data = output.data<float>();
  
  A2 = raw_data[0];
  A3 = raw_data[1];
  
  // return raw_data[1]-raw_data[0];

  return TRUE;
}
